#include <stdlib.h>
#include <string.h>
#include "enemy.h"
#include "map.h"
#include "player.h"
#include "screen.h"


/* Add enemy to our list of enemies */
static void add_enemy(enemy_t **enemies, int row, int col, char symbol)
{
    enemy_t *dude, *ee;

    dude = calloc(1, sizeof(enemy_t));
    dude->pos.row = row;
    dude->pos.col = col;
    dude->symbol = symbol;

    ee = *enemies;
    while (ee && ee->next)
      ee = ee->next;

    /* If this is the head */
    if (!ee)
      *enemies = dude;
    else
      ee->next = dude;
}


enemy_t *enemy_load_enemies(const map_t *map)
{
    int      i, j;
    enemy_t *enemies;

    enemies = NULL;
    for (i=0; i<map->n_rows; ++i)
      for (j=0; j<map->n_cols; ++j)
        if (IS_ENEMY(map->map[i][j]))
          add_enemy(&enemies, i, j, map->map[i][j]);

    return enemies;
}


void enemy_free_enemies(enemy_t *enemies)
{
    enemy_t *ee, *next;

    ee = enemies;

    while (ee)
    {
        next = ee->next;
        free(ee);
        ee = next;
    }
}


static void update_shooting(
    enemy_t     *enemy,
    player_t    *player,
    const map_t *map)
{
    int  did_hit;
    char something;

    if (getenv("DIOS_DISABLE_SHOOTING"))
      return;

    /* Haven't hit anything yet */
    did_hit = 0;

    /* Restore the map (clear the munitions trail) */
    map_restore_position(map, &enemy->shoot.pos);

    /* Advance the munition */
    map_adjust_position(&enemy->shoot.pos, enemy->shoot.dir);
    screen_render_char(enemy->shoot.symbol, &enemy->shoot.pos);

    /* See if the munition crashed into something */
    something = map->map[enemy->shoot.pos.row][enemy->shoot.pos.col];

    /* Hit wall (destroy munition) */
    if ((something == MAP_WALL) || (IS_ENEMY(something)))
      did_hit = 1;

    /* Hit player (damage player) */
    else if (player_check_collision(player, &enemy->shoot.pos) != -1)
    {
        did_hit = 1;
        player_loose_segment(player);
    }
    
    if (did_hit)
    {    
        enemy->state = ENEMY_IDLE;
        map_restore_position(map, &enemy->shoot.pos);
        memset(&enemy->shoot, 0, sizeof(shoot_t));
    }
}


static void update_idle(enemy_t *enemy)
{
    /* Decide to shoot or not next tick */
    if (rand() % 2)
    {
        /* Shoot in a pseudo random direction */
        enemy->shoot.pos = enemy->pos;
        enemy->shoot.dir = rand() % 4;
        enemy->shoot.symbol = '.';
        enemy->state = ENEMY_SHOOTING;
    }
    else
      enemy->state = ENEMY_IDLE;
}


/* Return direction if we can see the player and view is not obstructed */
static direction_t can_see_player(
    const enemy_t  *enemy,
    const player_t *player,
    const map_t    *map,
    direction_t     dir)
{
    int i, my_col, my_row;

    /* Adjust these to start searching from */
    my_row = enemy->pos.row;
    my_col = enemy->pos.col;

    /* Look through the map in given direction */
    if (dir == UP)
    {
        for (i=my_row-1; i>=0; --i)
        {
            if (map->map[i][my_col] != MAP_PATH)
              return NOWHERE;
            else if (player_check_collision(player,
                     &(const pos_t){i, my_col}) != -1)
              return dir;
        }
    }

    else if (dir == DOWN)
    {
        for (i=my_row+1; i<map->n_rows; ++i)
        {
            if (map->map[i][my_col] != MAP_PATH)
              return NOWHERE;
            else if (player_check_collision(player,
                     &(const pos_t){i, my_col}) != -1)
              return dir;
        }
    }
    
    else if (dir == LEFT)
    {
        for (i=my_col-1; i>=0; --i)
        {
            if (map->map[my_row][i] != MAP_PATH)
              return NOWHERE;
            else if (player_check_collision(player,
                     &(const pos_t){my_row, i}) != -1)
              return dir;
        }
    }
    
    else if (dir == RIGHT)
    {
        for (i=my_col+1; i<map->n_cols; ++i)
        {
            if (map->map[my_row][i] != MAP_PATH)
              return NOWHERE;
            else if (player_check_collision(player,
                     &(const pos_t){my_row, i}) != -1)
              return dir;
        }
    }

    return NOWHERE;
}


#define IS_SEEN(_dir) \
    ((seen_at = can_see_player(enemy, player, map, _dir)) != NOWHERE)
static void update_mover(enemy_t *enemy, player_t *player, const map_t *map)
{
    direction_t seen_at;

    /* Figure out where to move */
    if (IS_SEEN(UP) || IS_SEEN(DOWN) || IS_SEEN(LEFT) || IS_SEEN(RIGHT))
    {
        /* Clear the old position */
        map->map[enemy->pos.row][enemy->pos.col] = MAP_PATH;
        screen_render_char(MAP_PATH, &enemy->pos);

        /* Move the enemy (this will refresh so no need to call screen here) */
        map_adjust_position(&enemy->pos, seen_at);
        map->map[enemy->pos.row][enemy->pos.col] = enemy->symbol;

        /* If we are ontop of the player... eat em! */
        if (player_check_collision(player, &enemy->pos) != -1)
          player_loose_segment(player);
    }
}


void enemy_update(enemy_t *enemies, player_t *player, const map_t *map)
{
    enemy_t *ee;

    /* Check each enemy and decide to fire or not */
    for (ee=enemies; ee; ee=ee->next)
    {
        /* Dead dont dance */
        if (ee->state == ENEMY_DEAD)
          continue;

        /* If the enemy is not on the map, it probably has been shot */
        if (!IS_ENEMY(map->map[ee->pos.row][ee->pos.col]))
        {
            ee->state = ENEMY_DEAD;
            ee->symbol = MAP_PATH;
            continue;
        }
        
        switch (ee->state)
        {
            case ENEMY_SHOOTING: update_shooting(ee, player, map); break;
            case ENEMY_IDLE: update_idle(ee); break;
            default: break;
        }

        /* If we are a special kind of enemy (intelligent) */
        if (map->map[ee->pos.row][ee->pos.col] == MAP_ENEMY_MOVER)
          update_mover(ee, player, map);
    }
}
