#define _XOPEN_SOURCE 500
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <dirent.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include "enemy.h"
#include "main.h"
#include "map.h"
#include "player.h"
#include "screen.h"


static void usage(const char *execname)
{
    printf("===== " NAME " v" VER " =====\n"
           "Usage: %s <mapfile>\n", execname);

    exit(EXIT_SUCCESS);
}


static void handle_signal(int signum)
{
    /* Exit and reset the term */
    screen_exit();

    if (signum == SIGHUP || signum == SIGINT)
      exit(EXIT_SUCCESS);
    else
      exit(EXIT_FAILURE);
}


/* Play one round of the game (on specified map)
 * Auto play is the game playing itself.
 */
static game_status_t play_game(
    int         level_num,
    int         auto_play,
    const char *map_fname)
{
    int            tick;
    map_t         *map;
    time_t         start_time, now;
    enemy_t       *enemies;
    player_t      *player;
    game_status_t  game_status;

    /* Signals */
    signal(SIGHUP, handle_signal);
    signal(SIGINT, handle_signal);
    signal(SIGSEGV, handle_signal);

    /* Initalize game */
    game_status = GAME_ON;
    screen_display_map_name(level_num, map_fname);
    map = map_load_map(map_fname);
    player = player_create_player(map);
    enemies = enemy_load_enemies(map);

    /* Play game */
    tick = 0;
    start_time = time(NULL);
    while (game_status != GAME_OVER &&
           game_status != GAME_OVER_WIN)
    {
        /* Clock */
        now = time(NULL);
        
        /* Update screen */
        screen_render_map(map);
        screen_render_player(player);
        screen_render_enemies(enemies);
        screen_render_status(player, map, now - start_time);
        screen_refresh();

        /* If the player has been hit too many times (or timeout) */
        if ((player->icon_length == 0) ||
            ((now - start_time) >= TIMEOUT))
        {
            game_status = GAME_OVER;
            break;
        }

        /* Can we get no better? */
        else if (player->icon_length == PLAYER_MAX_ICONS)
        {
            game_status = GAME_OVER_WIN;
            break;
        }

        /* Update player */
        if (auto_play)
          game_status = player_auto_player(player, map);
        else
          game_status = player_move_player(player, map);
    
        /* Hurt player (or try to) */
        enemy_update(enemies, player, map);

        usleep(50000);
        ++tick;
    }

    /* Clean up and display results */
    map_free_map(map);
    player_free_player(player);
    enemy_free_enemies(enemies);
    screen_display_results(start_time, game_status);

    return game_status;
}


static char **get_map_list(const char *dirname, int *n_maps)
{
    int            idx, map_count, length;
    DIR           *dir;
    char          *c, **maps;
    struct dirent *entry;

    *n_maps = 0;

    if (!(dir = opendir(dirname)))
      ERR_FAIL("Could not open directory of maps: '%s'", dirname);

    /* Count files with .map extension */
    map_count = 0;
    while ((entry = readdir(dir)))
    {
        if (!(c = strrchr(entry->d_name, '.')) ||
            (strlen(c) != strlen(".map"))      ||
            (strncmp(c, ".map", 4) != 0))
          continue;

        ++map_count;
    }

    maps = malloc(sizeof(char *) * map_count);
    rewinddir(dir);

    idx = 0;
    while ((entry = readdir(dir)))
    {
        if (!(c = strrchr(entry->d_name, '.')) ||
            (strlen(c) != strlen(".map"))      ||
            (strncmp(c, ".map", 4) != 0))
          continue;
       
        length = strlen(entry->d_name) + strlen(dirname) + 2;
        maps[idx] = malloc(length);
        snprintf(maps[idx], length, "%s/%s", dirname, entry->d_name);
        ++idx;
    }

    *n_maps = map_count;
    return maps;
}


int main(int argc, char **argv)
{
    int             i, play_self, n_maps;
    char          **maps;
    const char     *map_path;
    struct stat     file_info;
    game_status_t   status;

    /* Defaults */
    play_self = 0;
    map_path = NULL;

    /* Args */
    for (i=1; i<argc; ++i)
      if (strncmp(argv[i], "-a", 2) == 0)
        play_self = 1;
      else
        map_path = argv[i];

    if (!map_path)
      usage(argv[0]);

    /* Initalize and display welcome */
    status = GAME_ON;
    srand(time(NULL));
    screen_init();
    screen_display_welcome();
        
    /* Determine if its a map (file) or a set of maps to play (directory) */
    stat(map_path, &file_info);

    /* Just play one level */
    if (!S_ISDIR(file_info.st_mode))
      play_game(1, play_self, map_path);

    /* Play all map files in the directory */
    else if (S_ISDIR(file_info.st_mode))
    {
        maps = get_map_list(map_path, &n_maps);
        for (i=0; i<n_maps && !GAME_LOSS(status); ++i)
        {
            status = play_game(i + 1, play_self, maps[i]);
            free(maps[i]);
        }
        free(maps);
    }

    /* Clean up */
    screen_exit();
    return EXIT_SUCCESS;
}
