#include <stdlib.h>
#include <string.h>
#include <ncurses.h>
#include "enemy.h"
#include "main.h"
#include "map.h"
#include "player.h"
#include "screen.h"


static void add_tail_segment(player_t *player)
{
    int idx;

    if ((player->icon_length + 1) > PLAYER_MAX_ICONS)
      return;

    ++player->icon_length;
    idx = player->icon_length - 1;
    player->icon[idx].pos = player->icon[idx-1].pos;

    /* Only two segments (tail and head) */
    if (player->icon_length == 2)
    {
        player->icon[0].symbol = 'D';
        player->icon[1].symbol = '8';
    }
    else
    {
        player->icon[idx].symbol = '8';
        player->icon[idx-1].symbol = '=';
    }
}


player_t *player_create_player(const map_t *map)
{
    player_t *pimp;

    pimp = calloc(1, sizeof(player_t));
    pimp->dir = NOWHERE;
    pimp->last_pos.row = map->start.row;
    pimp->last_pos.col = map->start.col;

    /* Head */
    pimp->icon[0].symbol = 'O';
    pimp->icon[0].pos.row = map->start.row;
    pimp->icon[0].pos.col = map->start.col;
    pimp->icon_length = 1;

    return pimp;
}


void player_free_player(player_t *player)
{
    free(player);
}


/* Returns boolean values if we can move up, down, left, right */
static int check_bounds(
    int             possible[N_DIRECTIONS], /* <-- returned */
    const player_t *player,
    const map_t    *map)
{
    int row, col, bound_row, bound_col;

    row = player->icon[0].pos.row;
    col = player->icon[0].pos.col;
    bound_row = map->n_rows;
    bound_col = map->n_cols;
    bound_row = map->n_rows;
    memset(possible, 0, N_DIRECTIONS);

    /* Up */
    if (((row-1) > 0) && (map->map[row-1][col] != MAP_WALL))
      possible[UP] = 1;
    
    /* Down */
    if (((row+1) < bound_row) && (map->map[row+1][col] != MAP_WALL))
      possible[DOWN] = 1;

    /* Left */
    if (((col-1) > 0) && (map->map[row][col-1] != MAP_WALL))
      possible[LEFT] = 1;

    /* Right */
    if (((col+1) < bound_col) && (map->map[row][col+1] != MAP_WALL))
      possible[RIGHT] = 1;

    return (possible[UP] + possible[DOWN] + possible[LEFT] + possible[RIGHT]);
}


/* Tail follows the previous icon location */
static void update_tail_and_map(player_t *player, const map_t *map)
{
    int i;
        
    /* No tail? */
    if (player->icon_length == 1)
    {
        /* Clear where we have been */
        map_restore_position(map, &player->last_pos);
        return;
    }
    
    /* Clear where we have been */
    map_restore_position(map, &player->icon[player->icon_length-1].pos);

    /* Just shift the positions from the end to the head */
    for (i=player->icon_length-1; i; --i)
      player->icon[i].pos = player->icon[i-1].pos;

    /* Adjust the first piece of the tail */
    player->icon[1].pos = player->last_pos;
}


/* Determine if we are on something, like a powerup */
static void check_if_on_something(player_t *player, const map_t *map)
{
    char something;

    something = map->map[player->icon[0].pos.row][player->icon[0].pos.col];

    if (something == MAP_POWER)
    {
        add_tail_segment(player);
        map_place_random(map, &player->icon[0].pos);
    }
    else if (IS_ENEMY(something))
      player_loose_segment(player);
}


static void update_shoot(player_t *player, const map_t *map)
{
    char something;

    if (!player->shoot.is_shooting)
      return;

    /* Restore the map (clear the munitions trail) */
    map_restore_position(map, &player->shoot.pos);

    /* Move the munition */
    map_adjust_position(&player->shoot.pos, player->shoot.dir);
    screen_render_char(player->shoot.symbol, &player->shoot.pos);

    /* Check to see if the munition has hit something */
    something = map->map[player->shoot.pos.row][player->shoot.pos.col];

    /* If we hit something */
    if ((something == MAP_WALL) || (IS_ENEMY(something)))
    {
        if (IS_ENEMY(something))
          map->map[player->shoot.pos.row][player->shoot.pos.col] = MAP_PATH;

        map_restore_position(map, &player->shoot.pos);
        player->shoot.is_shooting = 0;
    }
}


static void prepare_shoot(player_t *player, const map_t *map)
{
    /* If we have a shot launched, dont try to shoot an aditional */
    if (player->shoot.is_shooting)
      return;
    
    /* Initalize shot */ 
    player->shoot.pos = player->icon[0].pos;
    player->shoot.dir = player->dir;
    player->shoot.symbol = '~';
    player->shoot.is_shooting = 1;
}


/* Lastly, if player_auto_player or player_move_player is called 
 * This needs to be called to update and check for collisions
 */
static void update(player_t *player, const map_t *map)
{
    check_if_on_something(player, map);

    /* Let the munition go 2x than player */
    update_shoot(player, map);
    update_shoot(player, map);
}


/* Figure out where to move */
game_status_t player_auto_player(player_t *player, const map_t *map)
{
    int possible[N_DIRECTIONS] = {0}; /* Up, down, left, right */
    int i, n_possible, possible_try;

    /* Remember where we are now so we can have screen erase us */
    player->last_pos = player->icon[0].pos;

    /* Array of where we can move (up, down, left , right) */
    n_possible = check_bounds(possible, player, map);

    /* Choose pseudo-random position */
    possible_try = 1 + rand() % n_possible;

    /* Find the possilbe_idx position
     * We are using 4 instead of N_DIRECTIONS, since we have a NOWHERE
     * enumeration, and we don't want to check that.
     */
    for (i=0; i<4 && possible_try; ++i)
      if (possible[i])
        --possible_try;

    if (i != 0)
      --i;

    /* So we chose 'i' in our possible positions to move array */
    if (possible[i])
    {
        player->dir = (direction_t)i;
        map_adjust_position(&player->icon[0].pos, (direction_t)i);
        update_tail_and_map(player, map);
    }

    update(player, map);

    return GAME_ON;
}


game_status_t player_move_player(player_t *player, const map_t *map)
{
    int           key, possible[N_DIRECTIONS] = {0};
    direction_t   move_to;
    game_status_t status;

    /* Remember where we are now so we can have screen erase us */
    player->last_pos = player->icon[0].pos;

    status = GAME_ON;
    move_to = player->dir;
    check_bounds(possible, player, map);
    switch ((key = getch()))
    {
        case 'W':
        case 'w':
        case 'k':
        case KEY_UP: if (possible[UP]) move_to = UP; break;

        case 'S':
        case 's':
        case 'j':
        case KEY_DOWN: if (possible[DOWN]) move_to = DOWN; break;

        case 'A':
        case 'a':
        case 'h':
        case KEY_LEFT: if (possible[LEFT]) move_to = LEFT; break;

        case 'D':
        case 'd':
        case 'l':
        case KEY_RIGHT: if (possible[RIGHT]) move_to = RIGHT; break;

        case ' ': prepare_shoot(player, map); break;

        case ERR: break;

        default: return GAME_OVER;
    }

    /* If 'ERR' then we need to make sure its still within bounds */
    if (!possible[move_to])
      return GAME_ON;

    /* Move to our desired position */
    player->dir = move_to;
    map_adjust_position(&player->icon[0].pos, move_to);
    update_tail_and_map(player, map);

    update(player, map);

    return GAME_ON;
}


int player_check_collision(const player_t *player, const pos_t *pos)
{
    int i;

    for (i=0; i<player->icon_length; ++i)
      if (map_positions_equal(&player->icon[i].pos, pos))
        return i;

    return -1;
}


void player_loose_segment(player_t *player)
{
    /* No more segments to remove? */
    if (player->icon_length == 1)
      player->icon_length = 0;

    /* Last man left, UNO! */
    else if (player->icon_length == 2)
    {
        player->icon_length = 1;
        player->icon[0].symbol = 'O';
    }

    /* Remove tail, and add it one segment closer to head */
    else 
    {
        --player->icon_length;
        player->icon[player->icon_length - 1].symbol = '8';
    }
}
